/*
 * iconset.cpp - generic handling of iconsets
 * Copyright (C) 2003  Totemworks, LLC
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

//! \class Impix iconset.h
//! \brief Combines a QPixmap and QImage into one class
//!
//!  Normally, it is common to use QPixmap for all application graphics.
//!  However, sometimes it is necessary to access pixel data, which means
//!  a time-costly conversion to QImage.  Impix does this conversion on
//!  construction, and keeps a copy of both a QPixmap and QImage for fast
//!  access to each.  What you gain in speed you pay in memory, as an Impix
//!  should occupy twice the space.
//!
//!  An Impix can be conveniently created from either a QImage or QPixmap
//!  source, and can be casted back to either type.
//!
//!  You may also call unload() to free the image data.
//!
//!  \code
//!  Impix i = QImage("icon.png");
//!  QLabel *iconLabel;
//!  iconLabel->setPixmap(i.pixmap());
//!  QImage image = i.image();
//!  \endcode

#include"iconset.h"

#include<qpixmap.h>
#include<qimage.h>
#include<qptrvector.h>
#include<qdict.h>
#include<qmime.h>


//----------------------------------------------------------------------------
// Impix
//----------------------------------------------------------------------------
class Impix::Private
{
public:
	Private()
	{
		pixmap = 0;
		image = 0;
	}

	~Private()
	{
		unload();
	}

	void unload()
	{
		delete pixmap;
		pixmap = 0;
		delete image;
		image = 0;
	}

	QPixmap *pixmap;
	QImage *image;
};

//! \brief Construct a null Impix
//!
//! Constructs an Impix without any image data.
Impix::Impix()
{
	d = new Private;
}

//! \brief Construct an Impix based on a QPixmap
//!
//! Constructs an Impix by making a copy of a QPixmap and creating a QImage from it.
//! \param from - source QPixmap
Impix::Impix(const QPixmap &from)
{
	d = new Private;
	*this = from;
}

//! \brief Construct an Impix based on a QImage
//!
//! Constructs an Impix by making a copy of a QImage and creating a QPixmap from it.
//! \param from - source QImage
Impix::Impix(const QImage &from)
{
	d = new Private;
	*this = from;
}

//! \brief Copy constructor
Impix::Impix(const Impix &from)
{
	d = new Private;
	*this = from;
}

//! Sets the Impix to \param from
Impix & Impix::operator=(const Impix &from)
{
	d->unload();
	if(from.d->pixmap)
		d->pixmap = new QPixmap(*from.d->pixmap);
	if(from.d->image)
		d->image = new QImage(*from.d->image);
	return *this;
}

//! Unloads image data, making it null.
void Impix::unload()
{
	if(isNull())
		return;

	d->unload();
}

//! Destroys the image.
Impix::~Impix()
{
	delete d;
}

bool Impix::isNull() const
{
	return d->image ? true: false;
}

const QPixmap & Impix::pixmap() const
{
	if(!d->pixmap)
		d->pixmap = new QPixmap;
	return *d->pixmap;
}

const QImage & Impix::image() const
{
	if(!d->image)
		d->image = new QImage;
	return *d->image;
}

void Impix::setPixmap(const QPixmap &x)
{
	d->unload();
	d->pixmap = new QPixmap(x);
	d->image = new QImage(x.convertToImage());
}

void Impix::setImage(const QImage &x)
{
	d->unload();
	d->pixmap = new QPixmap;
	d->pixmap->convertFromImage(x);
	d->image = new QImage(x);
}


//----------------------------------------------------------------------------
// SimpleAnim
//----------------------------------------------------------------------------
SimpleAnim::SimpleAnim()
:QValueVector<Impix>()
{
}

SimpleAnim::~SimpleAnim()
{
}

int SimpleAnim::speed() const
{
	return v_speed;
}

void SimpleAnim::setSpeed(int x)
{
	v_speed = x;
}


//----------------------------------------------------------------------------
// Icon
//----------------------------------------------------------------------------
Icon::Icon(const Impix &icon)
{
	v_icon = icon;
	v_anim = 0;
}

Icon::Icon(const Impix &icon, const SimpleAnim &anim)
{
	v_icon = icon;
	setAnimation(anim);
}

Icon::~Icon()
{
	clearAnimation();
}

const Impix & Icon::icon() const
{
	return v_icon;
}

const SimpleAnim & Icon::animation() const
{
	return *v_anim;
}

bool Icon::isAnimated() const
{
	return v_anim ? true: false;
}

void Icon::setIcon(const Impix &icon)
{
	v_icon = icon;
}

void Icon::setAnimation(const SimpleAnim &anim)
{
	clearAnimation();
	v_anim = new SimpleAnim(anim);
}

void Icon::clearAnimation()
{
	delete v_anim;
	v_anim = 0;
}


//----------------------------------------------------------------------------
// Iconset
//----------------------------------------------------------------------------
class Iconset::Private
{
public:
	Private()
	{
		list.setAutoDelete(true);
	}

	QDict<Icon> list;
	Icon nullIcon;
};

Iconset::Iconset()
{
	d = new Private;
}

Iconset::Iconset(Iconset &from)
{
	d = new Private;
	*this = from;
}

Iconset & Iconset::operator=(Iconset &from)
{
	clear();
	QDictIterator<Icon> it(from.d->list);
	for(; it.current(); ++it)
		d->list.insert(it.currentKey(), new Icon(*it.current()));
	return *this;
}

Iconset::~Iconset()
{
	delete d;
}

const Icon & Iconset::icon(const QString &name) const
{
	if(d->list.isEmpty())
		return d->nullIcon;
	Icon *i = d->list.find(name);
	if(i)
		return *i;
	else
		return d->nullIcon;
}

void Iconset::setIcon(const QString &name, const Icon &icon)
{
	d->list.replace(name, new Icon(icon));
}

void Iconset::removeIcon(const QString &name)
{
	d->list.remove(name);
}

void Iconset::clear()
{
	d->list.clear();
}

QMimeSourceFactory *Iconset::createMimeSourceFactory() const
{
	QMimeSourceFactory *m = new QMimeSourceFactory;

	QDictIterator<Icon> it(d->list);
	for(; it.current(); ++it)
		m->setImage(it.currentKey(), it.current()->icon());
	return m;
}
