#include"filetransdlg.h"

#include<qlabel.h>
#include<qlineedit.h>
#include<qpushbutton.h>
#include<qtimer.h>
#include<qfiledialog.h>
#include<qfile.h>
#include<qprogressbar.h>
#include<qdir.h>
#include"psiaccount.h"
#include"userlist.h"
#include"common.h"
#include"iconwidget.h"
#include"busywidget.h"
#include"filetransfer.h"
#include"profiles.h"

#define CSMAX 64
#define CSMIN 16
static int calcShift(Q_LLONG big)
{
	Q_UINT64 val = 1;
	val <<= CSMAX - 1;
	for(int n = CSMAX - CSMIN; n > 0; --n) {
		if(big & val)
			return n;
		val >>= 1;
	}
	return 0;
}

//----------------------------------------------------------------------------
// FileTransferHandler
//----------------------------------------------------------------------------
class FileTransferHandler::Private
{
public:
	PsiAccount *pa;
	FileTransfer *ft;
	S5BConnection *c;
	Jid peer;
	QString fileName;
	Q_LLONG fileSize, sent;
	bool sending;
	QFile f;
	int shift;
};

FileTransferHandler::FileTransferHandler(PsiAccount *pa, FileTransfer *ft)
{
	d = new Private;
	d->pa = pa;
	d->c = 0;
	d->sent = 0;
	if(ft) {
		d->sending = false;
		d->peer = ft->peer();
		d->fileName = ft->fileName();
		d->fileSize = ft->fileSize();
		d->shift = calcShift(d->fileSize);
		d->ft = ft;
		Jid proxy = d->pa->userAccount().dtProxy;
		if(proxy.isValid())
			d->ft->setProxy(proxy);
		mapSignals();
	}
	else {
		d->sending = true;
		d->ft = 0;
	}
}

FileTransferHandler::~FileTransferHandler()
{
	if(d->ft) {
		d->ft->close();
		delete d->ft;
	}
	delete d;
}

void FileTransferHandler::send(const XMPP::Jid &to, const QString &fname)
{
	if(!d->sending)
		return;

	d->peer = to;
	QFileInfo fi(fname);
	d->fileName = fi.fileName();
	d->fileSize = fi.size(); // TODO: large file support
	d->shift = calcShift(d->fileSize);

	d->ft = d->pa->client()->fileTransferManager()->createTransfer();
	Jid proxy = d->pa->userAccount().dtProxy;
	if(proxy.isValid())
		d->ft->setProxy(proxy);
	mapSignals();

	d->f.setName(fname);
	d->ft->sendFile(d->peer, d->fileName, d->fileSize);
}

Jid FileTransferHandler::peer() const
{
	return d->peer;
}

QString FileTransferHandler::fileName() const
{
	return d->fileName;
}

Q_LLONG FileTransferHandler::fileSize() const
{
	return d->fileSize;
}

int FileTransferHandler::totalSteps() const
{
	return (d->fileSize >> d->shift);
}

void FileTransferHandler::accept(const QString &fname)
{
	if(d->sending)
		return;
	d->f.setName(fname);
	d->ft->accept();
}

void FileTransferHandler::s5b_proxyQuery()
{
	statusMessage(tr("Quering proxy..."));
}

void FileTransferHandler::s5b_proxyResult(bool b)
{
	if(b)
		statusMessage(tr("Proxy query successful."));
	else
		statusMessage(tr("Proxy query failed!"));
}

void FileTransferHandler::s5b_requesting()
{
	statusMessage(tr("Requesting data transfer channel..."));
}

void FileTransferHandler::s5b_accepted()
{
	statusMessage(tr("Peer accepted request."));
}

void FileTransferHandler::s5b_tryingHosts(const StreamHostList &)
{
	statusMessage(tr("Connecting to peer..."));
}

void FileTransferHandler::s5b_proxyConnect()
{
	statusMessage(tr("Connecting to proxy..."));
}

void FileTransferHandler::s5b_waitingForActivation()
{
	statusMessage(tr("Waiting for peer activation..."));
}

void FileTransferHandler::ft_accepted()
{
	d->c = d->ft->s5bConnection();
	connect(d->c, SIGNAL(proxyQuery()), SLOT(s5b_proxyQuery()));
	connect(d->c, SIGNAL(proxyResult(bool)), SLOT(s5b_proxyResult(bool)));
	connect(d->c, SIGNAL(requesting()), SLOT(s5b_requesting()));
	connect(d->c, SIGNAL(accepted()), SLOT(s5b_accepted()));
	connect(d->c, SIGNAL(tryingHosts(const StreamHostList &)), SLOT(s5b_tryingHosts(const StreamHostList &)));
	connect(d->c, SIGNAL(proxyConnect()), SLOT(s5b_proxyConnect()));
	connect(d->c, SIGNAL(waitingForActivation()), SLOT(s5b_waitingForActivation()));

	if(d->sending)
		accepted();
}

void FileTransferHandler::ft_connected()
{

	if(d->sending) {
		if(!d->f.open(IO_ReadOnly)) {
			delete d->ft;
			d->ft = 0;
			error(ErrFile, "");
			return;
		}

		trySend();
	}
	else {
		if(!d->f.open(IO_WriteOnly | IO_Truncate)) {
			delete d->ft;
			d->ft = 0;
			error(ErrFile, "");
			return;
		}
	}

	statusMessage(tr("Data transfer channel established."));
}

void FileTransferHandler::ft_readyRead(const QByteArray &a)
{
	if(!d->sending) {
		printf("%d bytes read\n", a.size());
		int r = d->f.writeBlock(a.data(), a.size());
		if(r < 0) {
			d->f.close();
			delete d->ft;
			d->ft = 0;
			error(ErrFile, "");
			return;
		}
		d->sent += a.size();
		if(d->sent == d->fileSize) {
			d->f.close();
			delete d->ft;
			d->ft = 0;
		}
		progress(d->sent >> d->shift, d->sent);
	}
}

void FileTransferHandler::ft_bytesWritten(int x)
{
	if(d->sending) {
		printf("%d bytes written\n", x);
		d->sent += x;
		if(d->sent == d->fileSize) {
			d->f.close();
			delete d->ft;
			d->ft = 0;
		}
		else
			QTimer::singleShot(0, this, SLOT(trySend()));
		progress(d->sent >> d->shift, d->sent);
	}
}

void FileTransferHandler::ft_error(int x)
{
	if(d->f.isOpen())
		d->f.close();
	delete d->ft;
	d->ft = 0;

	if(x == FileTransfer::ErrReject)
		error(ErrReject, "");
	else if(x == FileTransfer::ErrNeg)
		error(ErrTransfer, tr("Unable to negotiate data transfer channel."));
	else if(x == FileTransfer::ErrConnect)
		error(ErrTransfer, tr("Unable to connect to peer for data transfer."));
	else if(x == FileTransfer::ErrStream)
		error(ErrTransfer, tr("Lost connection."));
}

void FileTransferHandler::trySend()
{
	int blockSize = d->ft->dataSizeNeeded();
	QByteArray a(blockSize);
	int r = d->f.readBlock(a.data(), a.size());
	if(r < 0) {
		d->f.close();
		delete d->ft;
		d->ft = 0;
		error(ErrFile, "");
		return;
	}
	if(r < (int)a.size())
		a.resize(r);
	d->ft->writeFileData(a);
}

void FileTransferHandler::mapSignals()
{
	connect(d->ft, SIGNAL(accepted()), SLOT(ft_accepted()));
	connect(d->ft, SIGNAL(connected()), SLOT(ft_connected()));
	connect(d->ft, SIGNAL(readyRead(const QByteArray &)), SLOT(ft_readyRead(const QByteArray &)));
	connect(d->ft, SIGNAL(bytesWritten(int)), SLOT(ft_bytesWritten(int)));
	connect(d->ft, SIGNAL(error(int)), SLOT(ft_error(int)));
}

//----------------------------------------------------------------------------
// FileTransDlg
//----------------------------------------------------------------------------
static QString lastPath;

class FileTransDlg::Private
{
public:
	PsiAccount *pa;
	Jid jid;
	QLabel *lb_ident;
	FileTransferHandler *ft;
	QString fileName;
	Q_LLONG fileSize;
	bool sending;
	bool setTotal;
};

FileTransDlg::FileTransDlg(const Jid &jid, PsiAccount *pa)
:FileTransUI(0, 0, false, WDestructiveClose)
{
	d = new Private;
	d->pa = pa;
	d->jid = jid;
	d->ft = 0;
	d->sending = true;

	d->lb_ident = d->pa->accountLabel(this, true);
	d->lb_ident->setSizePolicy(QSizePolicy( QSizePolicy::Minimum, QSizePolicy::Minimum ));
	replaceWidget(lb_accountlabel, d->lb_ident);

	UserListItem *u = d->pa->findFirstRelavent(d->jid);
	setCaption(tr("Send File: %1").arg(jidnick(u->jid().full(), u->name())));

	le_to->setText(d->jid.full());
	le_to->setReadOnly(false);
	pb_start->setText(tr("&Send"));
	pb_stop->setText(tr("&Close"));

	connect(tb_browse, SIGNAL(clicked()), SLOT(chooseFile()));
	connect(pb_start, SIGNAL(clicked()), SLOT(doStart()));
	connect(pb_stop, SIGNAL(clicked()), SLOT(close()));

	lb_status->setText(tr("Ready"));

	pb_start->setFocus();
	d->pa->dialogRegister(this);
	QTimer::singleShot(0, this, SLOT(chooseFile()));
}

FileTransDlg::FileTransDlg(FileTransfer *ft, PsiAccount *pa)
:FileTransUI(0, 0, false, WDestructiveClose)
{
	d = new Private;
	d->pa = pa;
	d->jid = ft->peer();
	d->ft = new FileTransferHandler(pa, ft);
	d->sending = false;

	d->fileName = ft->fileName();
	d->fileSize = ft->fileSize();

	d->lb_ident = d->pa->accountLabel(this, true);
	d->lb_ident->setSizePolicy(QSizePolicy( QSizePolicy::Minimum, QSizePolicy::Minimum ));
	replaceWidget(lb_accountlabel, d->lb_ident);

	lb_to->setText(tr("From:"));
	UserListItem *u = d->pa->findFirstRelavent(d->jid);
	setCaption(tr("Receive File: %1").arg(jidnick(u->jid().full(), u->name())));

	QString fname = QDir::home().filePath(d->fileName);

	le_to->setText(u->jid().full());
	le_fname->setText(fname);
	lb_size->setText(tr("%1 byte(s)").arg(d->fileSize));
	pb_start->setText(tr("&Accept"));
	pb_stop->setText(tr("&Reject"));

	//connect(tb_browse, SIGNAL(clicked()), SLOT(chooseFile()));
	connect(pb_start, SIGNAL(clicked()), SLOT(doStart()));
	connect(pb_stop, SIGNAL(clicked()), SLOT(close()));

	lb_status->setText(tr("Ready"));

	pb_start->setFocus();
	d->pa->dialogRegister(this);
}

FileTransDlg::~FileTransDlg()
{
	delete d->ft;
	d->pa->dialogUnregister(this);
	delete d;
}

void FileTransDlg::done(int r)
{
	if(d->ft) {
		int n = QMessageBox::information(this, tr("Warning"), tr("Are you sure you want to cancel the transfer?"), tr("&Yes"), tr("&No"));
		if(n != 0)
			return;

		// close/reject FT if there is one
		if(d->ft) {
			delete d->ft;
			d->ft = 0;
		}
	}

	QDialog::done(r);
}

void FileTransDlg::chooseFile()
{
	if(lastPath.isEmpty())
		lastPath = QDir::homeDirPath();
	QString str = QFileDialog::getOpenFileName(lastPath, tr("All files (*.*)"), this, 0, tr("Choose a file"));
	if(!str.isEmpty()) {
		QFileInfo fi(str);
		lastPath = fi.dirPath();
		le_fname->setText(fi.filePath());
		lb_size->setText(tr("%1 byte(s)").arg(fi.size())); // TODO: large file support
	}
}

void FileTransDlg::doStart()
{
	Jid proxy = d->pa->userAccount().dtProxy;
	d->setTotal = true;

	if(d->sending) {
		pb_start->setEnabled(false);
		pb_stop->setText(tr("&Cancel"));
		pb_stop->setFocus();
		busy->start();

		QFileInfo fi(le_fname->text());
		d->fileName = fi.fileName();
		d->fileSize = fi.size(); // TODO: large file support

		d->ft = new FileTransferHandler(d->pa);
		connect(d->ft, SIGNAL(accepted()), SLOT(ft_accepted()));
		connect(d->ft, SIGNAL(statusMessage(const QString &)), SLOT(ft_statusMessage(const QString &)));
		connect(d->ft, SIGNAL(progress(int, Q_LLONG)), SLOT(ft_progress(int, Q_LLONG)));
		connect(d->ft, SIGNAL(error(int, const QString &)), SLOT(ft_error(int, const QString &)));
		d->ft->send(le_to->text(), le_fname->text());
	}
	else {
		QFileInfo fi(le_fname->text());
		if(fi.exists()) {
			QMessageBox::information(this, tr("Error"), tr("File already exists locally.  Cancelling transfer."));
			return;
		}

		pb_start->setEnabled(false);
		pb_stop->setText(tr("&Cancel"));
		pb_stop->setFocus();
		busy->start();

		connect(d->ft, SIGNAL(accepted()), SLOT(ft_accepted()));
		connect(d->ft, SIGNAL(statusMessage(const QString &)), SLOT(ft_statusMessage(const QString &)));
		connect(d->ft, SIGNAL(progress(int, Q_LLONG)), SLOT(ft_progress(int, Q_LLONG)));
		connect(d->ft, SIGNAL(error(int, const QString &)), SLOT(ft_error(int, const QString &)));
		d->ft->accept(le_fname->text());
	}
}

void FileTransDlg::ft_accepted()
{
	lb_status->setText(tr("Accepted!"));
}

void FileTransDlg::ft_statusMessage(const QString &s)
{
	lb_status->setText(s);
}

void FileTransDlg::ft_progress(int p, Q_LLONG sent)
{
	if(d->setTotal) {
		progressBar->setTotalSteps(d->ft->totalSteps());
		d->setTotal = false;
	}
	progressBar->setProgress(p);
	if(sent == d->fileSize) {
		delete d->ft;
		d->ft = 0;
		pb_stop->setText(tr("&Close"));
		busy->stop();
		lb_status->setText(tr("File transfer completed."));
		closeDialogs(this);
	}
}

void FileTransDlg::ft_error(int x, const QString &s)
{
	delete d->ft;
	d->ft = 0;
	pb_stop->setText(tr("&Close"));
	busy->stop();
	lb_status->setText(tr("File transfer failed."));
	closeDialogs(this);

	QString str;
	if(x == FileTransferHandler::ErrReject)
		str = tr("File was rejected by remote user.");
	else if(x == FileTransferHandler::ErrTransfer)
		str = tr("Error transferring file: %1").arg(s);
	else
		str = tr("File I/O error");
	QMessageBox::information(this, tr("Error"), s);
}
